

/* led.c
 * Routines for the LCD panels
 *
 * by Clemens Helfmeier
 * 5/2006
 */

#include "led.h"
#include <avr/io.h>
#include <avr/pgmspace.h>
#include <inttypes.h>
#include <avr/interrupt.h>


#define led_ActiveLow		1
#define led_ActiveHigh		2

/* define the polarity of the Segments. This could be
 * led_ActiveLow or led_ActiveHigh, whether the segment or digit
 * is on when the pin is low or when it is high
 */
#define led_SegmentPolarity		led_ActiveLow
#define led_DigitPolarity		led_ActiveLow

#define led_SegmentPort		PORTB
#define led_DigitPort		PORTD
#define led_SegmentDir		DDRB
#define led_DigitDir		DDRD


/* global variables */
/* this variable holds the value of the light
 * 0 means "off"
 * 1 and above means increasing illumination
 * 255 is maximum
 */
// led_Light is a macro that expands to TIMER0-CompareA-Register.
// uint8_t led_Light;

/* this is the counter for the currently displayed digit
 */
uint8_t led_Digit;

/* local variables */
uint8_t u8Data[led_Size] = { 0xFA, 0xFB, 0xFC, 0xFD, 0xFE };

/* these values select the segment on port led_SegmentPort
 */
#define DP		(1<<0)
#define RB		(1<<1)
#define MB		(1<<2)
#define LB		(1<<3)
#define MM		(1<<4)
#define LT		(1<<5)
#define MT		(1<<6)
#define RT		(1<<7)

/* this is the "character set"... */
PROGMEM uint8_t u8Segments[] = {
	RB | MB | LB | LT | MT | RT,		/* 0 */
	RB | RT,							/* 1 */
	MT | RT | MM | LB | MB,				/* 2 */
	MT | RT | MM | RB | MB,				/* 3 */
	LT | RT | MM | RB,					/* 4 */
	MT | LT | MM | RB | MB,				/* 5 */
	MT | LT | MM | LB | RB | MB,		/* 6 */
	MT | RT | RB,						/* 7 */
	MT | RT | LT | MM | LB | RB | MB,	/* 8 */
	MT | RT | LT | MM | RB | MB,		/* 9 */
	MT | RT | LT | MM | RB | LB,		/* A */
	LT | MM | LB | RB | MB,				/* b */
	MT | LT | LB | MB,					/* C */
	RT | MM | LB | RB | MB,				/* d */
	MT | LT | MM | LB | MB,				/* E */
	MT | LT | MM | LB,					/* F */
	MM,									/* - */
};

/* the following values will be used for selecting the digit
 * on port led_DigitPort
 */
#define D0		(1<<0)
#define D1		(1<<1)
#define D2		(1<<3)
#define D3		(1<<4)
#define D4		(1<<5)

PROGMEM uint8_t u8Digits[led_Size] = {
	D0,	/* digit 0 */
	D1, /* digit 1 */
	D2, /* digit 2 */
	D3, /* digit 3 */
	D4, /* digit 4 */
};

/* put the specified character (hex value) c to the display
 * at position pos.
 * returns 0 on success, any other value on failure
 */
uint8_t led_putHex( uint8_t pos, uint8_t c, uint8_t dp ) {
	if ((pos >= led_Size) || (c >= sizeof(u8Segments))) return -1;	/* out of range */
	uint8_t x = pgm_read_byte( &u8Segments[c] );
	if (dp != 0) x |= DP;
	u8Data[ pos ] = x;
	return 0;
}

uint8_t led_putDec( uint8_t pos, uint8_t i, uint8_t dp ) {
	if ((pos >= led_Size-1) || (i > 99)) return -1;	// out of range
	uint8_t x = pgm_read_byte( &u8Segments[i % 10] );
	if (dp == 1) x |= DP;
	u8Data[ pos ] = x;
	x = pgm_read_byte( &u8Segments[i / 10] );
	if (dp == 2) x |= DP;
	u8Data[ pos+1] = x;
	return 0;
}

void led_init( void ) {
	/* initialize the LEDs */
	/* configuure io-pins and turn off any segments */
#if led_SegmentPolarity == led_ActiveLow
	led_SegmentDir = 0x00;
	led_SegmentPort = 0x00;
#else
	led_SegmentDir = 0xFF;
	led_SegmentPort = 0x00;
#endif
#if led_DigitPolarity == led_ActiveLow
	led_DigitDir &= ~(D0 | D1 | D2 | D3 | D4);
	led_DigitPort &= ~(D0 | D1 | D2 | D3 | D4);
#else
	led_DigitDir |= D0 | D1 | D2 | D3 | D4;
	led_DigitPort &= ~(D0 | D1 | D2 | D3 | D4);
#endif
	led_setLight( 20000 );	/* full light */
	led_Digit = -1;		/* begin with first digit */

	/* inialize timer1 */
	// The timer runs to 20000 and starts back at 0 again. This results in
	// exactly 10 000 / 20 = 500Hz.
	// the OCR1A gives brightness of the display.
	TCCR1A = (0<<WGM11) | (0<<WGM10);
	TCCR1B = (1<<WGM13) | (1<<WGM12) | (0<<CS12) | (0<<CS11) | (1<<CS10);
	TCCR1C = 0;
	ICR1 = 20040;
	// enable the "overflow" and the compare interrupt
	TIMSK |= (1<<OCIE1A) | (1<<ICIE1);
}


void led_setLight( uint16_t light ) {
	if (light > 19000) light = 20100;
	OCR1A = light;
}

#if 0
void led_nextDigit( void ) {
#else
ISR(TIMER1_CAPT_vect) {
#endif
	register uint8_t p;
	/* turn off if not yet happend */
#if led_SegmentPolarity == led_ActiveLow
	led_SegmentDir = 0x00;
#else
	led_SegmentPort = 0x00;
#endif
	/* switch to the next digit */
#if led_DigitPolarity == led_ActiveLow
	p = led_DigitDir;
#else
	p = led_DigitPort;
#endif
	p &= ~(D0 | D1 | D2 | D3 | D4);
	led_Digit++;
	if (led_Digit >= led_Size) led_Digit = 0;
	p |= pgm_read_byte( &u8Digits[led_Digit] );
#if led_DigitPolarity == led_ActiveLow
	led_DigitDir = p;
#else	
	led_DigitPort = p;
#endif
#if led_SegmentPolarity == led_ActiveLow
	led_SegmentDir = u8Data[led_Digit];
#else
	led_SegmentPort = u8Data[led_Digit];
#endif
}

#if 0
void led_digitOff( void ) {
#else
ISR(TIMER1_COMPA_vect) {
#endif
	/* this only turns the light off */
#if led_SegmentPolarity == led_ActiveLow
	led_SegmentDir = 0x00;
#else
	led_SegmentPort = 0x00;
#endif
/*
#if led_DigitPolarity == led_ActiveLow
	led_DigitDir &= ~(D0 | D1 | D2 | D3 | D4);
#else
	led_DigitPort &= ~(D0 | D1 | D2 | D3 | D4)
#endif
*/
}
