

/** generator for dcf77-signals on TXD of Serial line.
 * This program sends the dcf77-stream to the TXD line of
 * a serial port.
 * The impulses are not exactly as long as they should be:
 *   a '0' should be 100ms ideally, but it is ((1+4)*1/50) =
 *      100ms
 *   a '1' should be 200ms long, it is ((1+8)*1/50) =
 *      180ms
 * most reciever will also work with these values. So this
 * program is useful for testing selfmade recievers if the
 * dcf77-signal is bad near TV sets or Computer monitors.
 *
 * The only possible parameter is the serial device. If not
 * given, the device /dev/ttyS1 is taken.
 * 
 * The Serial initialization stuff is taken from the Serial-HOWTO
 * by Gary Frerking which is found at
 * http://www.tldp.org/HOWTO/Serial-Programming-HOWTO/
 * 
 * @author Clemens Helfmeier
 * @date 11.8.2006
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <termios.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <time.h>

/* baudrate settings are defined in <asm/termbits.h>, which is
included by <termios.h> */
#define BAUDRATE B50
/* change this definition for the correct port */
char * device = "/dev/ttyS1";
#define _POSIX_SOURCE 1 /* POSIX compliant source */

#define FALSE 0
#define TRUE 1

void send_zero( int fd ) {
	// send a zero (short impuls)
	char v = 0xF0;
	write( fd, &v, 1 );
	printf("0");
	fflush(stdout);
}
void send_one( int fd ) {
	// send a one (long impuls)
	int v = 0x00;
	write( fd, &v, 1 );
	printf("1");
	fflush(stdout);
}

typedef struct {
	int len;
	int bits;
} sBitField;

int main( int argc, char * argv[]) {
	int fd,c, res;
	struct termios oldtio,newtio;
	char buf[255];

	if (argc == 2) {
		device = argv[1];
	}
	/*
	Open modem device for reading and writing and not as controlling tty
	because we don't want to get killed if linenoise sends CTRL-C.
	*/
	fd = open(device, O_RDWR | O_NOCTTY );
	if (fd <0) {perror(device); exit(-1); }

	tcgetattr(fd,&oldtio); /* save current serial port settings */
	bzero(&newtio, sizeof(newtio)); /* clear struct for new port settings */

	/*
	BAUDRATE: Set bps rate. You could also use cfsetispeed and cfsetospeed.
	CRTSCTS : output hardware flow control (only used if the cable has
	all necessary lines. See sect. 7 of Serial-HOWTO)
	CS8     : 8n1 (8bit,no parity,1 stopbit)
	CLOCAL  : local connection, no modem contol
	CREAD   : enable receiving characters
	*/
	newtio.c_cflag = BAUDRATE /*| CRTSCTS*/ | CS8 | CLOCAL | CREAD;

	/*
	IGNPAR  : ignore bytes with parity errors
	ICRNL   : map CR to NL (otherwise a CR input on the other computer
	will not terminate input)
	otherwise make device raw (no other input processing)
	*/
	newtio.c_iflag = IGNPAR | ICRNL;

	/*
	Raw output.
	*/
	newtio.c_oflag = 0;

	/*
	ICANON  : enable canonical input
	disable all echo functionality, and don't send signals to calling program
	*/
	newtio.c_lflag = ICANON;

	/*
	initialize all control characters
	default values can be found in /usr/include/termios.h, and are given
	in the comments, but we don't need them here
	*/
	newtio.c_cc[VINTR]    = 0;     /* Ctrl-c */
	newtio.c_cc[VQUIT]    = 0;     /* Ctrl-\ */
	newtio.c_cc[VERASE]   = 0;     /* del */
	newtio.c_cc[VKILL]    = 0;     /* @ */
	newtio.c_cc[VEOF]     = 4;     /* Ctrl-d */
	newtio.c_cc[VTIME]    = 0;     /* inter-character timer unused */
	newtio.c_cc[VMIN]     = 1;     /* blocking read until 1 character arrives */
	newtio.c_cc[VSWTC]    = 0;     /* '\0' */
	newtio.c_cc[VSTART]   = 0;     /* Ctrl-q */
	newtio.c_cc[VSTOP]    = 0;     /* Ctrl-s */
	newtio.c_cc[VSUSP]    = 0;     /* Ctrl-z */
	newtio.c_cc[VEOL]     = 0;     /* '\0' */
	newtio.c_cc[VREPRINT] = 0;     /* Ctrl-r */
	newtio.c_cc[VDISCARD] = 0;     /* Ctrl-u */
	newtio.c_cc[VWERASE]  = 0;     /* Ctrl-w */
	newtio.c_cc[VLNEXT]   = 0;     /* Ctrl-v */
	newtio.c_cc[VEOL2]    = 0;     /* '\0' */

	/*
	now clean the modem line and activate the settings for the port
	*/
	tcflush(fd, TCIFLUSH);
	tcsetattr(fd,TCSANOW,&newtio);

	/* generate the signal and output it */

#define PARITY( x ) ({char p = 0; for (int i = x; i!= 0; i=i>>1) p ^= i & 0x01; p;})
//(((x>>0) ^ (x>>1) ^ (x>>2) ^ (x>>3) ^ (x>>4) ^ (x>>5) ^ (x>>6) ^ (x>>7)) & 0x01)
#define hex2bcd( x ) ({ int p = 0, c = 0; for (int i = x; i!=0; i=i/10) { p |= (i % 10)<<c; c+=4; } p; })

	int streampos;
	while (1) {
		time_t at = time(0);
		struct tm * t = localtime( &at );
		if (t == 0) {
			printf( "Could not get local time.\n" );
			exit(-1);
		}
		sBitField stream[] = {
			{ 21, -2 },
			{ 7, hex2bcd(t->tm_min) },	/* minute */
			{ 1, PARITY(hex2bcd(t->tm_min))  }, /* parity 1 */
			{ 6, hex2bcd(t->tm_hour) }, /* hour */
			{ 1, PARITY(hex2bcd(t->tm_hour)) }, /* parity 2 */
			{ 6, hex2bcd(t->tm_yday) }, /* Kanlendertag */
			{ 3, hex2bcd(t->tm_wday) }, /* Wochentag */
			{ 5, hex2bcd(t->tm_mon) }, /* Kalendermonat */
			{ 8, hex2bcd(t->tm_year) }, /* Kalenderjahr */
			{ 1, PARITY(hex2bcd(t->tm_yday) ^ hex2bcd(t->tm_wday) ^ hex2bcd(t->tm_mon) ^ (hex2bcd(t->tm_year) & 0xFF)) }, /* parity 3 */
			{ 1, -1 }, /* SYNC */
			{ 0, 0 }, /* end of stream */
		};
		streampos = 0;
		while (stream[streampos].len != 0) {
			/* send the stream */
			int v = stream[streampos].bits;
			if (v == -1) {
				// send sync
				usleep(1000000);
				printf( "S\n" );
				usleep(1000000);
			} else {
				for (int i = 0; i<stream[streampos].len; i++) {
					if (v & 1)
						send_one( fd );
					else
						send_zero( fd );
					v = v >> 1;
					usleep(1000000);
				}
			}
			streampos++;
		}
	}

	
	
	/* restore the old port settings */
	tcsetattr(fd,TCSANOW,&oldtio);
}


